#pragma once
#include "fastapprox.h"
#include "pluginconstants.h"

#define EXTRACT_BITS(the_val, bits_start, bits_len) ((the_val >> (bits_start - 1)) & ((1 << bits_len) - 1))


/* paramToTimeCents()

	converts a time parameter (secs) to TimeCents
	See the MMA MIDI DLS Level 1 or 2 Spec

*/
inline double paramToTimeCents(double dParam)
{
	return 1200.0*65536.0*fastlog2(dParam);
}

/* timeCentsToParam()
	
	converts a timeCents to a time (sec) value
	See the MMA MIDI DLS Level 1 or 2 Spec

*/
inline double timeCentsToParam(double dTimeCents)
{
	return fastpow2(dTimeCents/(1200.0*65536.0));
}

/* mmaMIDItoParamScaling()
	
	calculates the scaled version of a parameter based on a MIDI value
	See the MMA MIDI DLS Level 1 or 2 Spec
	
	uMIDIValue = the MIDI (0 -> 127) value to apply
	fParameter = the parameter to scale
	fScaling = the scaling amount, 0 -> 1.0

	Examples:
		Velocity to EG Attack Time: attack time decreases as velocity increases
									uMIDIValue = velocity
									fParameter = attack time (mSec)
									fScaling = Vel to EG Attack Scale

		Key to EG Decay Time: decay shortens as MIDI note number increases
							  uMIDIValue = MIDI note number
							  fParameter = decay time (mSec)
							  fScaling = Vel to EG Attack Scale


*/
inline double mmaMIDItoParamScaling(UINT uMIDIValue, float fParameter, float fScaling)
{
	return timeCentsToParam(paramToTimeCents(fScaling)*(double(uMIDIValue)/128.0) + paramToTimeCents(fParameter));
}

/* mmaMIDItoAtten_dB()
	
	calculates the dB of attenuation according to MMA DLS spec
	
	uMIDIValue = the MIDI (0 -> 127) value to convert
*/
inline double mmaMIDItoAtten_dB(UINT uMIDIValue)
{
	if(uMIDIValue == 0)
		return -96.0; // dB floor

	return 20.0*log10((127.0*127.0)/((float)uMIDIValue*(float)uMIDIValue));
}

/* mmaMIDItoAtten()
	
	calculates the raw attenuation according to MMA DLS spec
	
	uMIDIValue = the MIDI (0 -> 127) value to convert
*/
inline double mmaMIDItoAtten(UINT uMIDIValue)
{
	if(uMIDIValue == 0)
		return 0.0; // floor

	return ((double)uMIDIValue*(double)uMIDIValue)/(127.0*127.0);;
}

/* convexTransform()
	
	calculates the convexTransform of the input
	
	dValue = the unipolar (0 -> 1) value to convert
*/
inline double convexTransform(double dValue)
{
	if(dValue <= 0)
		return 0.0;

	return 1.0 + (5.0/12.0)*log10(dValue);
}

/* convexInvertedTransform()
	
	calculates the convexInvertedTransform of the input
	
	dValue = the unipolar (0 -> 1) value to convert
*/
inline double convexInvertedTransform(double dValue)
{
	if(dValue <= 0)
		return 0.0;

	return 1.0 + (5.0/12.0)*log10(1.0 - dValue);
}

/* concaveTransform()
	
	calculates the concaveTransform of the input
	
	dValue = the unipolar (0 -> 1) value to convert
*/
inline double concaveTransform(double dValue)
{
	if(dValue >= 1.0)
		return 1.0;

	return -(5.0/12.0)*log10(1.0 - dValue);
}

/* concaveInvertedTransform()
	
	calculates the concaveInvertedTransform of the input
	
	dValue = the unipolar (0 -> 1) value to convert
*/
inline double concaveInvertedTransform(double dValue)
{
	if(dValue >= 1.0)
		return 1.0;

	return -(5.0/12.0)*log10(dValue);
}

/* unipolarToBipolar()
	
	calculates the bipolar (-1 -> +1) value from a unipolar (0 -> 1) value
	
	dValue = the value to convert
*/
inline double unipolarToBipolar(double dValue)
{
	return 2.0*dValue - 1.0;
}

/* bipolarToUnipolar()
	
	calculates the unipolar (0 -> 1) value from a bipolar (-1 -> +1) value
	
	dValue = the value to convert
*/
inline double bipolarToUnipolar(double dValue)
{
	return 0.5*dValue + 0.5;
}

/* calculatePanValues()
	
	calculates the left and right pan values
	
	dPanMod = unipolar (0 -> 1) pan modulation value

	returns are via pass-by-reference mechanism
*/
inline void calculatePanValues(double dPanMod, double& dLeftPanValue, double& dRightPanValue)
{
	// convert to BP
	dPanMod = unipolarToBipolar(dPanMod);

	dLeftPanValue = fastcos((pi/4.0)*(dPanMod + 1.0));
	dRightPanValue = fastsin((pi/4.0)*(dPanMod + 1.0));
}

/* calcValueVoltOctaveExp()
	
	returns the volt/octave exponential value
	
	dLowLimit = low limit of control
	dHighLimit = high limit of control
	dControlValue = on the range of 0 -> 1
*/
inline double calcValueVoltOctaveExp(double dLowLimit, double dHighLimit, double dControlValue)
{
	double dOctaves = fastlog2(dHighLimit/dLowLimit);
	if(dLowLimit == 0)
		return dControlValue;
	
	// exp control
	return dLowLimit*fastpow2(dControlValue*dOctaves);
}

/* calcInverseValueVoltOctaveExp()
	
	returns a value between 0 -> 1 representing the volt/octave control location
	
	dLowLimit = low limit of control
	dHighLimit = high limit of control
	dControlValue = value between low and high limit
*/
inline double calcInverseValueVoltOctaveExp(double dLowLimit, double dHighLimit, double dControlValue)
{
	double dOctaves = fastlog2(dHighLimit/dLowLimit);
	if(dLowLimit == 0)
		return dControlValue;
	
	return fastlog2(dControlValue/dLowLimit)/dOctaves;
}

/* calcModulatedValueExp()
	
	returns the new exponentially modulated control value
	
	dLowLimit = low limit of control
	dHighLimit = high limit of control
	dControlValue = current value of control
	dModValue = moudulator: on the range of 0 -> 1
	bExpUserControl = type of control on GUI; TRUE if user is adjusting a volt/octave control FALSE if linear
*/
inline double calcModulatedValueVoltOctaveExp(double dLowLimit, double dHighLimit, double dControlValue, 
											  double dModValue, bool bExpUserControl)
{
	dModValue = bipolarToUnipolar(dModValue);

	double dOctaves = fastlog2(dHighLimit/dLowLimit);
	
	// location of control relative to 50% point
	double dControlOffset = bExpUserControl ? calcInverseValueVoltOctaveExp(dLowLimit, dHighLimit,dControlValue) - 0.5 : ((dControlValue - dLowLimit)/(dHighLimit - dLowLimit)) - 0.5;

	// exp control
	double dExpControlOffset = dLowLimit*fastpow2(dControlOffset*dOctaves);
	
	// mod exp control
	double dModulatedValueExp = dExpControlOffset*fastpow2(dModValue*dOctaves);

	// bound
	dModulatedValueExp = min(dModulatedValueExp, dHighLimit);
	dModulatedValueExp = max(dModulatedValueExp, dLowLimit);

	// return it
	return dModulatedValueExp;
}

/* calcModulatedValueLin()
	
	returns the new linearly modulated control value
	
	dLowLimit = low limit of control
	dHighLimit = high limit of control
	dControlValue = current value of control
	dModValue = moudulator: on the range of 0 -> 1
*/
inline double calcModulatedValueLin(double dLowLimit, double dHighLimit, double dControlValue, double dModValue)
{	
	// convert mod value to bipolar
	double dModulator = unipolarToBipolar(dModValue)*(dHighLimit - dLowLimit)/2.0;// + dLowLimit; 
	
	double dModulatedValue = dControlValue + dModulator;

	// bound
	dModulatedValue = min(dModulatedValue, dHighLimit);
	dModulatedValue = max(dModulatedValue, dLowLimit);

	// return it
	return dModulatedValue;
}

inline double polyBLEP(double dModulo, double dInc, double dHeight, bool bRisingEdge, bool& bDone, double dEdgeLocation)
{
	// bDone is a flag letting the caller know we processed 
	// the point on the right side of the discontinuity
	bDone = false;

	// return value
	double dPolyBLEP = 0.0;

	// t = the distance from the discontinuity
	double t = 0.0;
	
	// LEFT side of edge
	// -1 < t < 0
	if((dEdgeLocation == 0 && dModulo > 1.0 - dInc) || 
		((dModulo > dEdgeLocation - dInc) && dModulo < dEdgeLocation))
	{ 
		// for the wrap-around edge only
		if(dEdgeLocation == 0.0)
			dEdgeLocation = 1.0;
		
		// calculate distance
		t = (dModulo - dEdgeLocation)/dInc; 

		// calculate residual
		dPolyBLEP = dHeight*(t*t + 2.0*t + 1.0); 
	
		// subtract for falling, add for rising edge
		if(!bRisingEdge)
			dPolyBLEP *= -1.0;

		return dPolyBLEP;
	} 

	// RIGHT side of discontinuity
	// 0 <= t < 1 
	if(dModulo < dEdgeLocation + dInc && dModulo > dEdgeLocation)
	{ 
		// set done flag
		bDone = true;

		// calculate distance
		t = (dModulo - dEdgeLocation)/dInc; 

		// calculate residual
		dPolyBLEP = dHeight*(2.0*t - t*t - 1.0); 
	
		// subtract for falling, add for rising edge
		if(!bRisingEdge)
			dPolyBLEP *= -1.0;

		return dPolyBLEP;
	} 

	return 0.0; 
} 

inline double doWhiteNoise()
{
	// fNoise is 0 -> 32767.0
	float fNoise = (float)rand();

	// normalize and make bipolar
	fNoise = 2.0*(fNoise/32767.0) - 1.0;

	return fNoise;
}

inline double doPNSequence(UINT& uPNRegister)
{
	// get the bits
	UINT b0 = EXTRACT_BITS(uPNRegister, 1, 1); // 1 = b0 is FIRST bit from right
	UINT b1 = EXTRACT_BITS(uPNRegister, 2, 1); // 1 = b1 is SECOND bit from right
	UINT b27 = EXTRACT_BITS(uPNRegister, 28, 1); // 28 = b27 is 28th bit from right
	UINT b28 = EXTRACT_BITS(uPNRegister, 29, 1); // 29 = b28 is 29th bit from right

	// form the XOR
	UINT b31 = b0^b1^b27^b28;

	// form the mask to OR with the register to load b31
	if(b31 == 1)
		b31 = 0x10000000;

	// shift one bit to right
	uPNRegister >>= 1; 

	// set the b31 bit
	uPNRegister |= b31;

	// convert the output into a floating point number, scaled by experimentation
	// to a range of o to +2.0
	float fOut = (float)(uPNRegister)/((pow((float)2.0,(float)32.0))/16.0);

	// shift down to form a result from -1.0 to +1.0
	fOut -= 1.0;

	return fOut;
}